import datetime
import json
import pickle

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Union, Dict

from appdata import FileStore


class KeyValueStore(ABC):

    @abstractmethod
    def flush(self, file_store: FileStore):
        pass

    @abstractmethod
    def init(self, file_store: FileStore):
        pass

    @abstractmethod
    def set_item(self, key: str, value: Any) -> None:
        pass

    @abstractmethod
    def get_item(self, key: str) -> Any:
        pass

    @abstractmethod
    def get_all(self) -> Dict:
        pass


class PickleStore(KeyValueStore):

    def __init__(self, path):
        super().__init__()
        self._path = path
        self._internal = {}

    def flush(self, file_store: FileStore) -> None:
        with file_store.file(self._path, 'wb') as f:
            pickle.dump(self._internal, f, protocol=pickle.HIGHEST_PROTOCOL)

    def init(self, file_store: FileStore) -> None:

        try:
            with file_store.file(self._path, 'rb') as f:
                self._internal = pickle.load(f)
        except FileNotFoundError:
            pass

    def set_item(self, key: str, value: Any) -> None:
        self._internal[key] = value

    def get_item(self, key: str) -> Any:
        return self._internal[key]

    def get_all(self) -> Dict:
        return self._internal.copy()


class JsonKVStore(KeyValueStore):

    def __init__(self, path: Union[str, Path]):
        self._path = path
        self._internal = {}

    def flush(self, file_store: FileStore):
        with file_store.file(self._path, 'w') as f:
            json.dump(self._internal, f, cls=_CustomEncoder)

    def init(self, file_store: FileStore):

        try:
            with file_store.file(self._path, 'r') as f:
                self._internal = json.load(f, object_hook=decode)
        except FileNotFoundError:
            pass

    def set_item(self, key: str, value: Any) -> None:
        self._internal[key] = value

    def get_item(self, key: str) -> Any:
        return self._internal[key]

    def get_all(self) -> Dict:
        return self._internal.copy()


class _CustomEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return dict(type="__datetime__", value=obj.isoformat())

        return super(_CustomEncoder, self).default(obj)


def decode(dct):
    if isinstance(dct, dict):
        if dct.get('type') == "__datetime__":
            return datetime.datetime.fromisoformat(dct["value"])
    return dct



from pathlib import Path
from typing import Any, Union, IO

from appdata.file_store import FileStore, Directory
from appdata.kv_store import KeyValueStore, PickleStore, JsonKVStore


class AppData:

    def __init__(self, key_value_store: Union[str, KeyValueStore] = "kv_store",
                 file_store: Union[str, FileStore] = "app_data",
                 auto_save=True
                 ):

        self._auto_save: bool = auto_save

        self._key_value_store: KeyValueStore = None
        self._file_store: FileStore = None

        self.set_file_store(file_store)
        self.set_key_value_store(key_value_store)

        self._kv_object = None

    def register(self, _kv_object):
        self._kv_object = _kv_object
        for key, val in self._key_value_store.get_all().items():
            setattr(_kv_object, key, val)

        def method(self, name, value, callback=self._set_to_store):
            callback(name, value)
            super(self.__class__, self).__setattr__(name, value)

        meths = {'__setattr__': method}
        _kv_object.__class__ = type('KV', (_kv_object.__class__,), meths)

    def _set_to_store(self, key, value):
        self._key_value_store.set_item(key, value)
        if self._auto_save:
            self.save()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()

    def __setitem__(self, key: str, value: Any):
        if self._kv_object is not None:
            setattr(self._kv_object, key, value)
        self._set_to_store(key, value)

    def __getitem__(self, key: str):
        return self._key_value_store.get_item(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def save(self):
        if self._file_store:
            self._key_value_store.flush(self._file_store)

    def set_key_value_store(self, key_value_store: KeyValueStore):
        if isinstance(key_value_store, str):
            self._key_value_store = JsonKVStore(key_value_store)
        else:
            self._key_value_store = key_value_store

        if self._file_store:
            self._key_value_store.init(self._file_store)

    def set_file_store(self, file_store: Union[str, FileStore]):
        if isinstance(file_store, str):
            self._file_store = Directory(file_store)
        else:
            self._file_store = file_store

    def write(self, path: Union[str, Path]) -> IO:
        return self._file_store.file(path, 'w')

    def read(self, path: Union[str, Path]) -> IO:
        return self._file_store.file(path, 'r')

    def write_binary(self, path: Union[str, Path]) -> IO:
        return self._file_store.file(path, 'wb')

    def read_binary(self, path: Union[str, Path]) -> IO:
        return self._file_store.file(path, 'rb')

    def file(self, path: Union[str, Path], mode) -> IO:
        return self._file_store.file(path, mode)


appdata = AppData()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['appdata']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'py-appdata',
    'version': '0.0.2',
    'description': 'Small utility library to provide easy access to files and key-value store in an application directory',
    'long_description': '# Python App Data\nThe py-appdata aims to solve the issue of storing simple application data and accessing it in an easy way.\n\nThe core idea is to expose a simple interface that has two properties\n - A get/set interface to simple values\n - A simple way of saving and retrieving files from the application directory\n\n## Stores\nThe `AppData` class provides an interface to two storage-abstractions: A file system and a key-value store. This lets\nyou easily save/retrieve files as well as setting/loading key/values which is a common app-data usage.\n\n## Get started\n````python\nfrom appdata import appdata\nfrom datetime import datetime, timezone\n\n# defaults to app_data\nappdata.set_file_store("my_app_data_dir")\n\n# defaults to kv_store\nappdata.set_key_value_store("my_store")\n\nappdata["last_login"] = datetime.utcnow().replace(tzinfo=timezone.utc)\n\nwith appdata.write("some_file.txt") as f:\n    f.write("Mjello")\n````\n\n## Dataclass support\nIf you like type hints, you can register a `dataclass` instance as a proxy for the key-value store\n```python\nfrom appdata import appdata\nfrom datetime import datetime, timezone\nfrom dataclasses import dataclass\n\n\n# Define the dataclass\n@dataclass\nclass KV:\n    last_login: datetime = None\n\n\n# Create an object to be referenced and register it to appdata\nkv = KV()\nappdata.register(kv)\n\n\n# The KeyValue object now acts as a proxy, and can be used throughout your project\nkv.last_login = datetime.utcnow().replace(tzinfo=timezone.utc)\n```',
    'author': 'Jesper Jensen',
    'author_email': 'mail@jeshj.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gedemagt/appdata',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



import os.path
from abc import ABC, abstractmethod
from pathlib import Path

from typing import Union, IO


class FileStore(ABC):

    @abstractmethod
    def exists(self, path: Union[str, Path]):
        pass

    @abstractmethod
    def file(self, path: Union[str, Path], mode='r') -> IO:
        pass


class Directory(FileStore):

    def __init__(self, path):
        self._parent = Path(path)

    def _to_path(self, path):
        return self._parent / path

    def exists(self, path: Union[str, Path]):
        return os.path.exists(self._parent / path)

    def file(self, path: Path, mode='r') -> IO:
        path = self._parent / path
        if 'w' in mode:
            path.parent.mkdir(parents=True, exist_ok=True)
        return open(path, mode)

    def __str__(self):
        return f"DirectoryStore<path={self._parent}>"
